﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using Microsoft.Concurrency.TestTools.UnitTesting;
using System.Threading.Algorithms;

namespace AlpacaProject
{
    /// <summary>
    /// This example derives from the EditDistance example from the following site:
    /// http://code.msdn.microsoft.com/ParExtSamples
    /// </summary>
    public class EditDistance
    {

        /// <summary>Generates a random string of lowercase chars.</summary>
        private static string GenerateRandomString(Random rand)
        {
            //const int LEN = 10000;
            const int LEN = 100;
            StringBuilder sb = new StringBuilder(LEN);
            for (int i = 0; i < LEN; i++)
                sb.Append((char)('a' + rand.Next(0, 26)));
            return sb.ToString();
        }

        TaskMeter serialMeter = new TaskMeter("Serial");
        TaskMeter wfNaiveMeter = new TaskMeter("Wavefront (naive)");
        TaskMeter wfBlockedMeter = new TaskMeter("Wavefront (blocked)");

        [PerformanceTestMethod(WarmupRepetitions = 2, Repetitions = 5)]
        public void CompareEditDistanceAlgorithms()
        {
            Random rand = new Random();
            string s1 = GenerateRandomString(rand);
            string s2 = GenerateRandomString(rand);
            long ticks;
            int parallelResult;

            serialMeter.Start();
            int serialResult = EditDistance_Serial(s1, s2);
            ticks = serialMeter.End();

            if (wfNaiveMeter.Enabled)
            {
                wfNaiveMeter.Start();
                parallelResult = EditDistance_Wavefront_Naive(s1, s2);
                ticks = wfNaiveMeter.End();
                Assert.AreEqual(serialResult, parallelResult, "EditDistance_Wavefront_Naive result");
            }

            if (wfBlockedMeter.Enabled)
            {
                wfBlockedMeter.Start();
                parallelResult = EditDistance_Wavefront_Blocked(s1, s2);
                ticks = wfBlockedMeter.End();
                Assert.AreEqual(serialResult, parallelResult, "EditDistance_Wavefront_Blocked result");
            }

            //GC.Collect();
        }

        private static int EditDistance_Serial(string s1, string s2)
        {
            // Initialize the 2D array of intermediate distances
            int[,] dist = new int[s1.Length + 1, s2.Length + 1];
            for (int i = 0; i <= s1.Length; i++) dist[i, 0] = i;
            for (int j = 0; j <= s2.Length; j++) dist[0, j] = j;

            for (int i = 1; i <= s1.Length; i++)
            {
                for (int j = 1; j <= s2.Length; j++)
                {
                    GetElementDistance(s1, s2, dist, i, j);
                }
            }
            return dist[s1.Length, s2.Length];
        }

        private static int EditDistance_Wavefront_Naive(string s1, string s2)
        {
            // Initialize the 2D array of intermediate distances
            int[,] dist = new int[s1.Length + 1, s2.Length + 1];
            for (int i = 0; i <= s1.Length; i++) dist[i, 0] = i;
            for (int j = 0; j <= s2.Length; j++) dist[0, j] = j;

            int numBlocks = Environment.ProcessorCount * 4;

            ParallelAlgorithms.Wavefront(s1.Length, s2.Length, (i, j) => {
                if (i == 0)
                    dist[i, j] = j;
                else if (j == 0)
                    dist[i, j] = i;
                else
                    GetElementDistance(s1, s2, dist, i, j);
            });

            return dist[s1.Length, s2.Length];
        }

        private static int EditDistance_Wavefront_Blocked(string s1, string s2)
        {
            // Initialize the 2D array of intermediate distances
            int[,] dist = new int[s1.Length + 1, s2.Length + 1];
            for (int i = 0; i <= s1.Length; i++) dist[i, 0] = i;
            for (int j = 0; j <= s2.Length; j++) dist[0, j] = j;

            int numBlocks = Environment.ProcessorCount * 4;

            ParallelAlgorithms.Wavefront(s1.Length, s2.Length, numBlocks, numBlocks, (start_i, end_i, start_j, end_j) => {
                for (int i = start_i + 1; i <= end_i; i++)
                {
                    for (int j = start_j + 1; j <= end_j; j++)
                    {
                        GetElementDistance(s1, s2, dist, i, j);
                    }
                }
            });

            return dist[s1.Length, s2.Length];
        }

        /// <summary>
        /// Performs the Dynamic-Programming part for single element in the 2D distance array.
        /// </summary>
        /// <returns></returns>
        private static int GetElementDistance(string s1, string s2, int[,] dist, int i, int j)
        {
            if (s1[i - 1] == s2[j - 1])
            {
                // The chars are the same, so the distance is the same
                return dist[i - 1, j - 1];
            }
            else
            {
                /*
                 * 1 + MIN(dist[i-1, j]
                 *       , dist[i, j-1]
                 *       , dist[i-1, j-1]
                 *       )
                 */
                return 1 + Math.Min(
                    dist[i - 1, j], Math.Min(
                    dist[i, j - 1],
                    dist[i - 1, j - 1]));
            }
        }

    }
}
