﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AlpacaProject
{
    /// <summary>
    /// This test class implements the strategies mentioned in the paper
    /// Patterns of Parallel Programming (PoPP) by Stephen Toub
    /// Section: ONE TASK PER ELEMENT
    /// </summary>
    /// <remarks>
    /// http://www.microsoft.com/downloads/details.aspx?FamilyID=86b3d32b-ad26-4bb8-a3ae-c1637026c3ee&displaylang=en
    /// </remarks>
    public static class MyParallel
    {

        public delegate T[] InvokeFuncDelegate<T>(params Func<T>[] functions);

        public static void InvokeAction_OneTaskPerElement(params Action[] actions)
        {
            var tasks = new Task[actions.Length];
            for (int i = 0; i < actions.Length; i++)
            {
                tasks[i] = Task.Factory.StartNew(actions[i]);
            }
            Task.WaitAll(tasks);
        }

        // Approach #1: One Task per element
        public static T[] InvokeFunc_OneTaskPerElement<T>(params Func<T>[] functions)
        {
            var tasks = (from function in functions
                         select Task.Factory.StartNew(function))
                         .ToArray();
            Task.WaitAll(tasks);
            return tasks.Select(t => t.Result).ToArray();
        }

        // Approach #2: One Task per element, using parent/child Relationships
        public static T[] InvokeFunc_ParentChildTasks<T>(params Func<T>[] functions)
        {
            var results = new T[functions.Length];
            Task.Factory.StartNew(() => {
                for (int i = 0; i < functions.Length; i++)
                {
                    int cur = i;
                    Task.Factory.StartNew(
                        () => results[cur] = functions[cur](),
                        TaskCreationOptions.AttachedToParent);
                }
            }).Wait();
            return results;
        }

        // Approach #3: Using Parallel.For
        public static T[] InvokeFunc_ParallelFor<T>(params Func<T>[] functions)
        {
            T[] results = new T[functions.Length];
            Parallel.For(0, functions.Length, i => {
                results[i] = functions[i]();
            });
            return results;
        }

        // Approach #4: Using PLINQ
        public static T[] InvokeFunc_Plinq<T>(params Func<T>[] functions)
        {
            return functions.AsParallel().Select(f => f()).ToArray();
        }

        public static T[] InvokeFunc_LimitDegreeOfParallelism<T>(params Func<T>[] functions)
        {
            T[] results = new T[functions.Length];
            var options = new ParallelOptions { MaxDegreeOfParallelism = 2 };
            Parallel.For(0, functions.Length, options, i => {
                results[i] = functions[i]();
            });
            return results;
        }


    }
}
