﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Concurrency.TestTools.UnitTesting;
using System.Threading;
using System.Threading.Tasks;

namespace AlpacaProject
{
    /// <summary>
    /// This test class implements the strategies mentioned in the paper
    /// Patterns of Parallel Programming (PoPP) by Stephen Toub
    /// Section: CLOSING OVER INAPPROPRIATELY SHARED DATA
    /// </summary>
    /// <remarks>
    /// http://www.microsoft.com/downloads/details.aspx?FamilyID=86b3d32b-ad26-4bb8-a3ae-c1637026c3ee&displaylang=en
    /// </remarks>
    public class PoPP_ClosingOverSharedData
    {

        /// <summary>
        /// What do you think this code will/should print out?
        /// </summary>
        [UnitTestMethod]
        public void SharedVarInWorkerItems()
        {
            for (int i = 0; i < 10; i++)
            {
                ThreadPool.QueueUserWorkItem(_ => Console.WriteLine(i));
            }
        }

        [UnitTestMethod]
        public void SequentialExecution()
        {
            var actions = new List<Action>();
            for (int i = 0; i < 10; i++)
                actions.Add(() => Console.WriteLine(i));
            actions.ForEach(action => action());
        }

        [UnitTestMethod]
        public void SharedVarInWorkerItems_Fixed()
        {
            // The var i is shared by both this method and each thread created by the ThreadPool.
            for (int i = 0; i < 10; i++)
            {
                // Use a loop-local variable to allow the closure to use the loop's current iteration value
                int iLoopLocal = i;
                ThreadPool.QueueUserWorkItem(delegate { Console.WriteLine(iLoopLocal); });
            }
        }

        [UnitTestMethod]
        public void SequentialExecution_Fixed()
        {
            var actions = new List<Action>();
            // The var i is shared by both the loop and each action.
            for (int i = 0; i < 10; i++)
            {
                // Fix: Use a loop-local variable to allow the closure to use the loop's current iteration value
                int iLoopLocal = i;
                actions.Add(() => Console.WriteLine(iLoopLocal));
            }
            actions.ForEach(action => action());
        }

        [UnitTestMethod]
        [RegressionTestExpectedResult(TestResultType.Exception)]
        public void VarDeclaredAtTopOfMethod()
        {
            // This is often done out of habit from C coding styles.
            int j;

            Parallel.For(0, 100, i => {
                int total = 0;
                for (j = 1; j <= 10000; j++)
                    total += j;
                Console.WriteLine("i={0}; total={0}", i, total);
                Assert.AreEqual(50005000, total);   // This will faile for any thread but the first.
            });
        }

        [UnitTestMethod]
        public void VarDeclaredAtTopOfMethod_Fixed()
        {
            Parallel.For(0, 100, i => {
                int total = 0;
                // Fix: Move declarations as close to their usage as possible
                for (int j = 1; j <= 10000; j++)
                    total += j;
                Console.WriteLine("total={0}", total);
                Assert.AreEqual(50005000, total);
            });
        }

    }
}
