﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.ParallelComputingPlatform.ParallelExtensions.Samples;
using System.Threading.Tasks;
using System.Collections.Concurrent;

namespace AlpacaProject.RayTracing
{
    class FineVsCoarseGrainedRayTracer : BasicParallelRaytracer
    {

        public enum RenderMethod
        {
            Sequential,
            ParallelFor,
            NestedParallelFor,
            FlattenedLoop,
            Partitioned,
        }

        public FineVsCoarseGrainedRayTracer(int screenWidth, int screenHeight)
            : base(screenWidth, screenHeight)
        {
        }

        public RenderMethod Method { get; set; }

        public override void Render(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            switch (Method)
            {
                case RenderMethod.Sequential:
                    Render_Sequential(scene, rgbBuffer, parallelOpts);
                    break;

                case RenderMethod.ParallelFor:
                    Render_ParallelFor(scene, rgbBuffer, parallelOpts);
                    break;

                case RenderMethod.NestedParallelFor:
                    Render_NestedParallelFor(scene, rgbBuffer, parallelOpts);
                    break;

                case RenderMethod.FlattenedLoop:
                    Render_FlattenedLoop(scene, rgbBuffer, parallelOpts);
                    break;

                case RenderMethod.Partitioned:
                    Render_Partitioned(scene, rgbBuffer, parallelOpts);
                    break;

                default: throw new NotImplementedException();
            }
        }

        private void Render_Sequential(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            for (int y = 0; y < screenHeight; y++)
            {
                int stride = y * screenWidth;
                for (int x = 0; x < screenWidth; x++)
                {
                    Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                    Color color = TraceRay(ray, scene, 0);
                    rgbBuffer[x + stride] = color.ToInt32();
                }
            }
        }

        private void Render_ParallelFor(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            Parallel.For(0, screenHeight, parallelOpts, body: y => {
                int stride = y * screenWidth;
                for (int x = 0; x < screenWidth; x++)
                {
                    Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                    Color color = TraceRay(ray, scene, 0);
                    rgbBuffer[x + stride] = color.ToInt32();
                }
            });
        }

        private void Render_NestedParallelFor(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            Parallel.For(0, screenHeight, parallelOpts, body: y => {
                int stride = y * screenWidth;
                //for (int x = 0; x < screenWidth; x++)
                Parallel.For(0, screenWidth, parallelOpts, body: x => {
                    Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                    Color color = TraceRay(ray, scene, 0);
                    rgbBuffer[x + stride] = color.ToInt32();
                });
            });
        }

        private void Render_FlattenedLoop(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            int totalPixels = screenWidth * screenHeight;

            Parallel.For(0, totalPixels, parallelOpts, body: i => {
                int x = i % screenWidth;
                int y = i / screenWidth;
                Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                Color color = TraceRay(ray, scene, 0);
                rgbBuffer[i] = color.ToInt32();
            });
        }

        private void Render_Partitioned(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            int totalPixels = screenWidth * screenHeight;

            Parallel.ForEach(Partitioner.Create(0, totalPixels), parallelOpts, body: range => {
                for (int i = range.Item1; i < range.Item2; i++)
                {
                    int x = i % screenWidth;
                    int y = i / screenWidth;
                    Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                    Color color = TraceRay(ray, scene, 0);
                    rgbBuffer[i] = color.ToInt32();
                }
            });
        }

    }
}
