﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Microsoft Corporation.  All rights reserved. 
// 
//  File: Raytracer.cs
//
//--------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace Microsoft.ParallelComputingPlatform.ParallelExtensions.Samples
{
    public class BasicParallelRaytracer : ParallelRayTracerBase
    {

        public BasicParallelRaytracer(int screenWidth, int screenHeight)
            : base(screenWidth, screenHeight)
        {
        }

        public bool ShowThreads { get; set; }

        public override void Render(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            if (ShowThreads)
                RenderShowingThreads(scene, rgbBuffer, parallelOpts);
            else
                RenderNoShowThreads(scene, rgbBuffer, parallelOpts);
        }

        private void RenderNoShowThreads(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            Parallel.For(0, screenHeight, parallelOpts, body: y => {
                int stride = y * screenWidth;
                for (int x = 0; x < screenWidth; x++)
                {
                    Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                    Color color = TraceRay(ray, scene, 0);
                    rgbBuffer[x + stride] = color.ToInt32();
                }
            });
        }

        private void RenderShowingThreads(Scene scene, Int32[] rgbBuffer, ParallelOptions parallelOpts)
        {
            int id = 0;
            Camera camera = scene.Camera;
            Vector cameraPos = camera.Pos;
            Parallel.For<double>(0, screenHeight, parallelOpts
                // Local Init
                , () => GetHueShift(Interlocked.Increment(ref id))
                // Body
                , (y, state, hue) => {
                    int stride = y * screenWidth;
                    for (int x = 0; x < screenWidth; x++)
                    {
                        Ray ray = new Ray(cameraPos, GetPoint(x, y, camera));
                        Color color = TraceRay(ray, scene, 0);
                        color.ChangeHue(hue);
                        rgbBuffer[x + stride] = color.ToInt32();
                    }
                    return hue;
                }
                // local finally
                , hue => Interlocked.Decrement(ref id)
                );
        }

    }
}
